from typing import Optional, Union, TYPE_CHECKING, overload, Dict

if TYPE_CHECKING:
    from pyexlatex.presentation.beamer.overlay.overlay import Overlay
    from pyexlatex.models.package import Package
    from pyexlatex.typing import PyexlatexItem

from copy import deepcopy
from mixins.attrequals import EqOnAttrsMixin, EqHashMixin
from pyexlatex.models.mixins import StringAdditionMixin, IsSpecificClassMixin
from pyexlatex.texgen import (
    _basic_item_str,
    _multi_option_item_str,
    _no_braces_item_str,
    item_equals_str,
    no_options_no_contents_str
)



class IsLatexItemMixin:
    is_LatexItem = True


class DataItem:
    _data_has_been_initialized = False

    def init_data(self):
        from pyexlatex.models.documentsetup import DocumentSetupData
        if not self._data_has_been_initialized:
            self._data_has_been_initialized = True
            self.data = DocumentSetupData()

    def add_package(self, package: Union[str, 'Package']):
        self.init_data()
        self.data.packages.append(package)


class ItemBase(DataItem, IsSpecificClassMixin, IsLatexItemMixin, StringAdditionMixin, EqHashMixin, EqOnAttrsMixin):
    """
    Base class for all latex generating classes
    
    Note: Does not pass any args to super call, so do not put another class below this one expecting args to
    be passed
    """
    

    def __init__(self, *args, **kwargs):
        self.init_data()
        super().__init__()

    @overload
    def _wrap_with(self, item: 'PyexlatexItem', begin_wrap: str, end_wrap: str,
                   format_contents: bool = True) -> str:
        ...
    @overload
    def _wrap_with(self, item: None, begin_wrap: str, end_wrap: str,
                   format_contents: bool = True) -> None:
        ...
    def _wrap_with(self, item, begin_wrap, end_wrap, format_contents=True):
        if format_contents:
            fmt = self._format_content
        else:
            fmt = lambda x: x

        if item is None:
            return None

        return f'{begin_wrap}{fmt(item)}{end_wrap}'

    @overload
    def _wrap_with_bracket(self, item: 'PyexlatexItem') -> str:
        ...
    @overload
    def _wrap_with_bracket(self, item: None) -> None:
        ...
    def _wrap_with_bracket(self, item):
        return self._wrap_with(item, '[', ']')

    @overload
    def _wrap_with_braces(self, item: 'PyexlatexItem') -> str:
        ...
    @overload
    def _wrap_with_braces(self, item: None) -> None:
        ...
    def _wrap_with_braces(self, item):
        return self._wrap_with(item, '{', '}')

    @staticmethod
    def _dict_to_options_str(d: Dict[str, 'PyexlatexItem']) -> str:
        return ','.join([f'{key}={value}' for key, value in d.items()])

    @staticmethod
    def _empty_str_if_none(item: Optional[str]) -> str:
        return item if item is not None else ''

    @staticmethod
    def _format_content(content):
        from pyexlatex.logic.format.contents import format_contents as fmt
        return fmt(content)

    @staticmethod
    def _get_list_copy_from_list_or_none(list_or_none: Optional[list]) -> list:
        if list_or_none is None:
            return []
        return deepcopy(list_or_none)




class Item(ItemBase):
    equal_attrs = [
        'contents',
        'pre_env_contents',
        'post_env_contents',
        'data',
        'env'
    ]

    def __init__(self, name, contents, pre_env_contents=None, post_env_contents=None, env_modifiers=None,
                 overlay: Optional['Overlay'] = None):
        from pyexlatex.models.environment import Environment
        self.env = Environment(name, modifiers=env_modifiers, overlay=overlay)
        self.contents = contents
        self.pre_env_contents = pre_env_contents
        self.post_env_contents = post_env_contents
        self.overlay = overlay
        super().__init__()

    def __repr__(self):
        return f'<Item(name={self.env.name}, contents={self.contents})>'

    def __str__(self):
        from pyexlatex.logic.builder import _build, build

        contents = deepcopy(self.contents)
        contents = build(contents)

        possible_items = [
            self.pre_env_contents,
            self.env.wrap(str(contents)),
            self.post_env_contents
        ]
        items = [item for item in possible_items if item]
        return _build(items)


class SimpleItem(ItemBase):
    equal_attrs = [
        'contents',
        'modifiers',
        'pre_modifiers',
        'overlay',
        'format_content',
        'data'
    ]

    def __init__(self, name, contents, modifiers: Optional[str] = None, pre_modifiers: Optional[str] = None,
                 overlay: Optional['Overlay'] = None, format_content: bool = True):
        self.name = name
        self.contents = contents
        self.modifiers = modifiers
        self.pre_modifiers = pre_modifiers
        self.overlay = overlay
        self.format_content = format_content
        super().__init__()

    def __repr__(self):
        return f'<{self.name.title()}({self.contents})>'

    def __str__(self):
        return _basic_item_str(
            self.name,
            self.contents,
            self.modifiers,
            self.pre_modifiers,
            overlay=self.overlay,
            format_content=self.format_content
        )


class MultiOptionSimpleItem(ItemBase):

    def __init__(self, name, *options, overlay: Optional['Overlay'] = None):
        self.name = name
        self.options = options
        self.overlay = overlay
        super().__init__()

    def __repr__(self):
        return f'<{self.name.title()}({self.options})>'

    def __str__(self):
        return _multi_option_item_str(self.name, *self.options, overlay=self.overlay)


class NoBracesItem(ItemBase):

    def __init__(self, name, contents, overlay: Optional['Overlay'] = None):
        self.name = name
        self.contents = contents
        self.overlay = overlay
        super().__init__()

    def __repr__(self):
        return f'<{self.name.title()}({self.contents})>'

    def __str__(self):
        return _no_braces_item_str(self.name, self.contents, overlay=self.overlay)


class EqualsItem(ItemBase):

    def __init__(self, name, contents):
        self.name = name
        self.contents = contents
        super().__init__()

    def __repr__(self):
        return f'<{self.name.title()}({self.contents})>'

    def __str__(self):
        return item_equals_str(self.name, self.contents)


class NoOptionsNoContentsItem(ItemBase):

    def __init__(self, name, overlay: Optional['Overlay'] = None, modifiers: Optional[str] = None):
        self.name = name
        self.overlay = overlay
        self.modifiers = modifiers
        super().__init__()

    def __repr__(self):
        return f'<{self.name.title()}>'

    def __str__(self):
        return no_options_no_contents_str(self.name, overlay=self.overlay, modifiers=self.modifiers)


from typing import Sequence, Optional

from pyexlatex.models.section.base import TextAreaBase


class BeamerColorBox(TextAreaBase):
    name = 'beamercolorbox'

    def __init__(self, content, color: str, options: Optional[Sequence[str]] = None, **kwargs):
        self.color = color
        self.options = options
        super().__init__(self.name, content, env_modifiers=self.modifier_str, **kwargs)

    @property
    def modifier_str(self) -> str:
        if self.options is None:
            return self._wrap_with_braces(self.color)
        options_str = ', '.join(self.options)
        return f'[{options_str}]{{{self.color}}}'



from typing import Type

from pyexlatex.models.section.sections import SubSubSection, SubSection, Section, Chapter


class TextSectionTest:
    item_class: Type = Section
    tag_name: str = 'section'

    def test_str(self):
        content = self.item_class('woo', title='title')
        braces_tag_name = f'{{{self.tag_name}}}'
        assert str(content) == f'\\begin{braces_tag_name}{{title}}\nwoo\n\end{braces_tag_name}'


class TestSubSubSection(TextSectionTest):
    item_class = SubSubSection
    tag_name = 'subsubsection'


class TestSubSection(TextSectionTest):
    item_class = SubSection
    tag_name = 'subsection'


class TestSection(TextSectionTest):
    item_class = Section
    tag_name = 'section'


class TestChapter(TextSectionTest):
    item_class = Chapter
    tag_name = 'chapter'


def test_sections_from_dict():
    contents = {
        'S One': 'Content for S One',
        'S Two': {
            'SS One': 'Content for SS One',
            'SS Two': {
                'SSS One': 'Content for SSS One',
            }
        },
    }
    content = Chapter(contents, title='C One')
    assert str(content) == '\\begin{chapter}{C One}\n\\begin{section}{S One}\nContent for S One\n\\end{section}\n\\begin{section}{S Two}\n\\begin{subsection}{SS One}\nContent for SS One\n\\end{subsection}\n\\begin{subsection}{SS Two}\n\\begin{subsubsection}{SSS One}\nContent for SSS One\n\\end{subsubsection}\n\\end{subsection}\n\\end{section}\n\\end{chapter}'



import math
import os

from pyexlatex.tools import date_time_move_latex


def latex_equations_to_pdf(latex_list, directory, name='Equations', below_text=None,
                           math_size=18, text_size=14, title=None, para_space='1em',
                          inline=False, as_document=True):
    script_size = math.ceil(math_size * (2/3))
    scriptscript_size = math.ceil(math_size * .5)
    assert text_size in (8, 9, 10, 11, 12, 14, 17, 20) #latex allowed font sizes

    if inline:
        surround_char_beg = '$'
        surround_char_end = '$'
    else:
        surround_char_beg = r'\begin{dmath}'
        surround_char_end = r'\end{dmath}'

    # Header and footer are needed to create a standalone document using the equations.
    # If as_document=False, header and footer will not be used.
    headers = [r'\documentclass[{} pt]{{extarticle}}'.format(text_size),
               #First size is text size, second is math size, third is script size,
               #fourth is scriptscript size
               r'\DeclareMathSizes{{{0}}}{{{1}}}{{{2}}}{{{3}}}'.format(
                    text_size, math_size, script_size, scriptscript_size),
               r'\usepackage{amsmath}',
               r'\usepackage{breqn}',
               r'\usepackage[margin=0.3in]{geometry}',
              r'\author{Nick DeRobertis}' ,r'\begin{document}', r'\setlength{{\parskip}}{{{}}}'.format(para_space)]
    footers = [r'\end{document}']

    name_tex = name + '.tex'
    file_path = os.path.join(directory, name_tex)

    # Actually write to file
    with open(file_path, 'w') as f:
        if as_document:
            f.write('\n'.join(headers) + '\n')
        [f.write(surround_char_beg + '{}'.format(line) + surround_char_end + '\n\n') for line in latex_list]
        if below_text:
            f.write('\n' + below_text + '\n')
        if as_document:
            f.write('\n'.join(footers))

    os.chdir(directory)
    # Only create pdf if we are creating a standalone document
    if as_document:
        os.system('pdflatex ' + '"' + name_tex + '"') #create pdf
    date_time_move_latex(name, directory, 'Equations')


from typing import Union, Sequence

from pyexlatex.models.commands.renewcommand import ReNewCommand
from pyexlatex.models.item import SimpleItem


class RightFooter(SimpleItem):
    """
    A right-aligned footer
    """
    name = 'rfoot'

    def __init__(self, contents):
        super().__init__(self.name, contents)


class CenterFooter(SimpleItem):
    """
    A center-aligned footer
    """
    name = 'cfoot'

    def __init__(self, contents):
        super().__init__(self.name, contents)


class LeftFooter(SimpleItem):
    """
    A left-aligned footer
    """
    name = 'lfoot'

    def __init__(self, contents):
        super().__init__(self.name, contents)


class FooterLine(ReNewCommand):
    """
    Add a horizontal line above the footer
    """

    def __init__(self, size: float = 0.4, units: str = 'pt'):
        size_str = f'{size}{units}'
        super().__init__(r'footrulewidth', size_str)


FooterType = Union[RightFooter, LeftFooter, CenterFooter, FooterLine]
Footers = Sequence[FooterType]



from typing import Iterable

class RowAddMixin:

    def __add__(self, other):
        return RowAddMixin.add(self, other)

    def __radd__(self, other):
        return RowAddMixin.radd(self, other)

    def _add_class(self, other):
        from pyexlatex.table.models.table.row import Row
        # keep same class if both are same class
        # otherwise, default to Row class
        self_class = type(self)
        other_class = type(other)
        klass = self_class if self_class == other_class else Row

        return klass

    @staticmethod
    def add(obj, other):
        klass = obj._add_class(other)

        return _add_to_value_or_values(obj, other, klass)

    @staticmethod
    def radd(obj, other):
        klass = obj._add_class(other)

        return _radd_to_value_or_values(obj, other, klass)

def _add_to_value_or_values(obj, other, klass):
    if hasattr(obj, 'value'):
        return _add_to_value(obj, other, klass)
    if hasattr(obj, 'values'):
        return _add_to_values(obj.values, other, klass)
    else:
        raise NotImplementedError

def _radd_to_value_or_values(obj, other, klass):
    if hasattr(obj, 'value'):
        return _radd_to_value(obj, other, klass)
    if hasattr(obj, 'values'):
        return _radd_to_values(obj.values, other, klass)
    else:
        raise NotImplementedError

def _add_to_value(value, other, klass):
    # handle named classes which have value or values attr
    if hasattr(other, 'value'):
        return klass([value, other.value])
    if hasattr(other, 'values'):
        return klass([value] + other.values)

    # handle builtin classes.
    # add lists, tuples, etc
    if isinstance(other, Iterable):
        return klass([value] + list(other))
    # treat as single item, create list out of the two items
    else:
        return klass([value, other])


def _add_to_values(values, other, klass):
    # handle named classes which have value or values attr
    if hasattr(other, 'value'):
        return klass(values + [other.value])
    if hasattr(other, 'values'):
        return klass(values + other.values)

    # handle builtin classes.
    # add lists, tuples, etc
    if isinstance(other, Iterable):
        return klass(values + list(other))
    # treat as single item, append to existing list
    else:
        return klass(values + [other])

def _radd_to_value(value, other, klass):
    # handle builtin classes.
    # add lists, tuples, etc
    if isinstance(other, Iterable):
        return klass(list(other) + [value])
    # treat as single item, create list out of the two items
    else:
        return klass([other, value])


def _radd_to_values(values, other, klass):
    # handle builtin classes.
    # add lists, tuples, etc
    if isinstance(other, Iterable):
        return klass(list(other) + values)
    # treat as single item, append to existing list
    else:
        return klass([other] + values)



from typing import List
import bibtexparser
from bibtexparser.bibdatabase import BibDatabase
from pyexlatex.models.references.bibtex.generic import BibTexEntry


def extract_bibtex_str(bibtex: str) -> List[BibTexEntry]:
    bib_db: BibDatabase = bibtexparser.loads(bibtex)
    bibtex_entries = [BibTexEntry(entry) for entry in bib_db.entries]
    return bibtex_entries



from pyexlatex.models.package import Package
from pyexlatex.texgen.packages.columntypes import ColumnTypesPackage

_default_package_strs = [
    'amsmath',
    'pdflscape',
    'booktabs',
    'array',
    'threeparttable',
    'fancyhdr',
    'lastpage',
    'textcomp'
]

_direct_default_packages = [
    ColumnTypesPackage(),
    Package('fontenc', modifier_str='T1'),  # allows printing of > and <
]

default_packages = [Package(package_str) for package_str in _default_package_strs] + _direct_default_packages  # type: ignore


from typing import Optional
from pyexlatex.models.item import SimpleItem


class Title(SimpleItem):
    name = 'title'

    def __init__(self, main_title: str, short_title: Optional[str] = None):
        self.main_title = main_title
        self.short_title = short_title
        super().__init__(self.name, main_title, pre_modifiers=self._wrap_with_bracket(short_title))


import os
import shutil


def _move_if_needed(src, dst):
    try:
        shutil.move(src, dst)
    except shutil.SameFileError:
        pass


def _move_if_exists_and_is_needed(src, dst):
    if not os.path.exists(src):
        return

    _move_if_needed(src, dst)


def _move_folder_or_move_files_if_destination_folder_exists(src, dst):
    try:
        _move_if_needed(src, dst)
    except shutil.Error:
        files = [file for file in next(os.walk(src))[2]]
        [_move_if_needed(file, dst) for file in files]


def _copy_if_needed(src, dst):
    try:
        shutil.copy(src, dst)
    except shutil.SameFileError:
        pass


from typing import Optional, Sequence, Union, List
import pandas as pd
from pyexlatex.models.item import Item
from pyexlatex.models.containeritem import ContainerItem
from mixins.repr import ReprMixin
from pyexlatex.table.models.panels.collection import PanelCollection
from pyexlatex.table.models.texgen.tabularbase import BaseTabular
from pyexlatex.table.models.texgen.alignment import ColumnsAlignment
from pyexlatex.table.logic.table.build import build_tabular_content_from_panel_collection
from pyexlatex.models.caption import Caption
from pyexlatex.texgen import _centering_str
from pyexlatex.models.document import Document
from pyexlatex.models.package import Package
from pyexlatex.models.landscape import Landscape
from pyexlatex.models.label import Label
from pyexlatex.models.section.base import TextAreaBase
from pyexlatex.table.models.data.table import DataTable


class TableNotes(TextAreaBase, ReprMixin):
    name = 'tablenotes'

    def __init__(self, contents: Union[str, Sequence[str]]):
        super().__init__(self.name, contents, env_modifiers=f'[para, flushleft]')


class Tabular(ContainerItem, Item, ReprMixin, BaseTabular):
    """
    The main part of a LaTeX table which contains the rows and columns.
    """
    name = 'tabular'
    repr_cols = ['align']

    def __init__(self, content, align: Optional[Union[ColumnsAlignment, str]] = None):
        if not isinstance(content, (list, tuple)):
            content = [content]
        self.align = self._get_columns_alignment_from_passed_align(content, align)
        self.add_package('booktabs')  # Remove this once booktabs is added in cmidrule and others
        self.add_data_from_content(self.align)
        super().__init__(self.name, content, env_modifiers=self._wrap_with_braces(str(self.align)))

    @classmethod
    def from_panel_collection(cls, panel_collection: PanelCollection, align: Optional[ColumnsAlignment] = None,
                 mid_rules=True):
        align = align if align is not None else ColumnsAlignment(num_columns=panel_collection.num_columns)
        obj = cls([[0]], align=align)  # dummy content
        obj.panel_collection = panel_collection
        obj.contents = build_tabular_content_from_panel_collection(panel_collection, mid_rule=mid_rules)

        return obj

    @classmethod
    def from_df(cls, df: pd.DataFrame, align: Optional[ColumnsAlignment] = None, **dt_from_df_kwargs):
        dt = DataTable.from_df(df, **dt_from_df_kwargs)
        return cls(dt, align=align)


class ThreePartTable(TextAreaBase, ReprMixin):
    name = 'threeparttable'
    repr_cols = ['caption']

    def __init__(self, table_content: Tabular, caption: Caption=None, below_text: TableNotes=None,
                 label: Optional[Label] = None):
        self.caption = caption
        items = [
            caption,
            table_content,
            below_text,
            label
        ]
        content = [item for item in items if item is not None]

        super().__init__(self.name, content)

    @classmethod
    def from_panel_collection(cls, panel_collection: PanelCollection, *args, tabular_kwargs={}, **kwargs):
        tabular = Tabular.from_panel_collection(panel_collection, **tabular_kwargs)

        caption: Optional[Caption]
        if panel_collection.name is not None:
            caption = Caption(panel_collection.name)
        else:
            caption = None

        return cls(tabular, caption=caption, **kwargs)

class Table(TextAreaBase, ReprMixin):
    name = 'table'
    repr_cols = ['caption']

    def __init__(self, three_part_table: ThreePartTable, centering=True, landscape=False):
        self.caption = three_part_table.caption
        self.landscape = landscape

        items = [
            _centering_str() if centering else None,
            three_part_table
        ]

        content = [item for item in items if item is not None]

        super().__init__(self.name, content)

    def __str__(self):
        content_with_env = super().__str__()
        if self.landscape:
            content_with_env = Landscape().wrap(str(content_with_env))
        return content_with_env

    @classmethod
    def from_panel_collection(cls, panel_collection: PanelCollection, *args, tabular_kwargs={},
                              three_part_table_kwargs={}, **kwargs):
        three_part_table = ThreePartTable.from_panel_collection(
            panel_collection,
            tabular_kwargs=tabular_kwargs,
            **three_part_table_kwargs
        )
        return cls(three_part_table, *args, **kwargs)

    @classmethod
    def from_table_model(cls, table, *args, **kwargs):
        from pyexlatex.table.models.table.table import Table as TableModel
        table: TableModel
        tabular = Tabular.from_panel_collection(
            table.panels,
            align=table.align,
            mid_rules=table.mid_rules
        )

        three_part_table = ThreePartTable(
            tabular,
            caption=table.caption,
            below_text=table.below_text,
            label=table.label
        )
        obj = cls(three_part_table, *args, landscape=table.landscape, **kwargs)
        obj.add_data_from_content(table)
        return obj



class LTable(Table):
    name = 'ltable'

    def __str__(self):
        # Skip landscape wrapping being done in Table as it is already handled by ltable definition
        return TextAreaBase.__str__(self)


class TableDocument(Document):

    def __init__(self, content: Table, packages: List[Package]=None, landscape: bool=False):
        from pyexlatex.table.models.texgen.packages import default_packages

        if packages is None:
            packages = []

        packages += default_packages

        super().__init__(content, packages, landscape=landscape)

    @classmethod
    def from_panel_collection(cls, panel_collection: PanelCollection, *args, tabular_kwargs={},
                              three_part_table_kwargs={}, table_kwargs={}, **kwargs):
        table = Table.from_panel_collection(
            panel_collection,
            tabular_kwargs=tabular_kwargs,
            three_part_table_kwargs=three_part_table_kwargs,
            **table_kwargs
        )
        return cls(table, *args, **kwargs)

    @classmethod
    def from_table_model(cls, table, *args, **kwargs):
        from pyexlatex.table.models.table.table import Table as TableModel
        table: TableModel
        tex_table = Table.from_table_model(table, *args, **kwargs)
        obj = cls(tex_table, *args, **kwargs)
        obj.add_data_from_content(table)
        return obj



"""
Add exceptions here and set _match_str to whatever string should match the exception. They will automatically
be passed to the exception manager
"""
import sys
import inspect
from pyexlatex.logic.output.errors.models.exception_manager import LatexExceptionManager


class LatexException(Exception):
    _match_str = ''


class TooManyUnprocessedFloatsException(LatexException):
    _match_str = 'Too many unprocessed floats'


class OutputLoopConsecutiveDeadCycles(LatexException):
    _match_str = 'consecutive dead cycles'

####### Exception Manager Logic (only define specific exceptions above here) #############

exception_classes =  [class_tuple[1] for class_tuple in inspect.getmembers(sys.modules[__name__], inspect.isclass)]
excluded_classes = [
    LatexException,
    LatexExceptionManager
]
specific_exception_classes = [klass for klass in exception_classes if klass not in excluded_classes]
exception_manager = LatexExceptionManager(specific_exception_classes)


from sympy import Eq, symbols
import pyexlatex as pl

STR_EQ = r'a = b^{2} + \frac{c}{d}'
a, b, c, d = symbols('a b c d')
SYMPY_EQ = Eq(a, b ** 2 + c/d)


def test_eq_inline():
    eq_from_str = pl.Equation(str_eq=STR_EQ)
    eq_from_sympy = pl.Equation(eq=SYMPY_EQ)

    assert str(eq_from_str) == str(eq_from_sympy) == f'${STR_EQ}$'


def test_eq_offset_numbered():
    eq_from_str = pl.Equation(str_eq=STR_EQ, inline=False, numbered=True)
    eq_from_sympy = pl.Equation(eq=SYMPY_EQ, inline=False, numbered=True)

    assert str(eq_from_str) == str(eq_from_sympy) == '\\begin{equation}\n\t' + STR_EQ + '\n\\end{equation}'


def test_eq_offset_no_numbers():
    eq_from_str = pl.Equation(str_eq=STR_EQ, inline=False, numbered=False)
    eq_from_sympy = pl.Equation(eq=SYMPY_EQ, inline=False, numbered=False)

    assert str(eq_from_str) == str(eq_from_sympy) == f'\n$${STR_EQ}$$\n'



from typing import Iterable


def _max_len_or_zero(iterable: Iterable) -> int:
    try:
        return max([len(i) for i in iterable])
    except ValueError:
        return 0


def _readable_repr(repr_str):
    out_letters = []
    num_tabs = 1
    for letter in repr_str:
        if letter in (')',']'):
            num_tabs -= 1
            out_letters += ['\n'] + ['   '] * num_tabs
        out_letters.append(letter)
        if letter in ('(','['):
            out_letters += ['\n'] + ['   '] * num_tabs
            num_tabs += 1
    return ''.join(out_letters)


def _add_if_not_none(*items):
    not_none_items = [item for item in items if item is not None]

    if len(not_none_items) == 0:
        return None
    elif len(not_none_items) == 1:
        return not_none_items[0]

    return sum(not_none_items[1:], not_none_items[0])


from pyexlatex.models.control.documentclass.classtypes.documentclasstype import DocumentClassType

RESUME_DEFINITION = r"""
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Medium Length Professional CV - RESUME CLASS FILE
%
% This template has been downloaded from:
% http://www.LaTeXTemplates.com
%
% This class file defines the structure and design of the template. 
%
% Original header:
% Copyright (C) 2010 by Trey Hunner
%
% Copying and distribution of this file, with or without modification,
% are permitted in any medium without royalty provided the copyright
% notice and this notice are preserved. This file is offered as-is,
% without any warranty.
%
% Created by Trey Hunner and modified by www.LaTeXTemplates.com
%
% Updated by Nick DeRobertis
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

\ProvidesClass{resume}[2010/07/10 v0.9 Resume class]

\LoadClass[11pt,letterpaper]{article} % Font size and paper type

\usepackage[parfill]{parskip} % Remove paragraph indentation
\usepackage{array} % Required for boldface (\bf and \bfseries) tabular columns
\usepackage{ifthen} % Required for ifthenelse statements

\pagestyle{empty} % Suppress page numbers

% Required for avoiding page breaks immediately after section titles
\usepackage{needspace}
\usepackage[explicit,pagestyles]{titlesec}

%----------------------------------------------------------------------------------------
%	HEADINGS COMMANDS: Commands for printing name and address
%----------------------------------------------------------------------------------------

\def \name#1{\def\@name{#1}} % Defines the \name command to set name
\def \@name {} % Sets \@name to empty by default

\def \addressSep {$\diamond$} % Set default address separator to a diamond

% One, two or three address lines can be specified 
\let \@addressone \relax
\let \@addresstwo \relax
\let \@addressthree \relax

% \address command can be used to set the first, second, and third address (last 2 optional)
\def \address #1{
  \@ifundefined{@addressone}{
    \def \@addressone {#1}
  }{
  \@ifundefined{@addresstwo}{
  \def \@addresstwo {#1}
  }{
     \@ifundefined{@addressthree}{
      \def \@addressthree {#1}
    }{}
  }}
}

% \printaddress is used to style an address line (given as input)
\def \printaddress #1{
  \begingroup
    \def \\ {\addressSep\ }
    \centerline{#1}
  \endgroup
  \par
  \addressskip
}

% \printname is used to print the name as a page header
\def \printname {
  \begingroup
    \hfil{\MakeUppercase{\namesize\bf \@name}}\hfil
    \nameskip\break
  \endgroup
}

%----------------------------------------------------------------------------------------
%	PRINT THE HEADING LINES
%----------------------------------------------------------------------------------------

\let\ori@document=\document
\renewcommand{\document}{
  \ori@document  % Begin document
  \printname % Print the name specified with \name
  \@ifundefined{@addressone}{}{ % Print the first address if specified
    \printaddress{\@addressone}}
  \@ifundefined{@addresstwo}{}{ % Print the second address if specified
    \printaddress{\@addresstwo}}
     \@ifundefined{@addressthree}{}{ % Print the third address if specified
    \printaddress{\@addressthree}}
}

%----------------------------------------------------------------------------------------
%	SECTION FORMATTING
%----------------------------------------------------------------------------------------

% Defines the ResumeSection environment for the large sections within the CV
\renewenvironment{section}[1]{ % 1 input argument - section name
  \needspace{10\baselineskip}
  \sectionskip
  \MakeUppercase{\bf #1} % Section title
  \sectionlineskip
  \hrule % Horizontal line
}

%----------------------------------------------------------------------------------------
%	WORK EXPERIENCE FORMATTING
%----------------------------------------------------------------------------------------

\newenvironment{employment}[4]{ % 4 input arguments - company name, year(s) employed, job title and location
 {\bf #1} \hfill {#2} % Bold company name and date on the right
 \ifthenelse{\equal{#3}{}}{}{ % If the third argument is not specified, don't print the job title and location line
  \\
  {\em #3} \hfill {\em #4} \\[-12pt] % Italic job title and location
  }\smallskip
  \begin{list}{$\cdot$}{\leftmargin=1em} % \cdot used for bullets, no indentation
   \itemsep -0.5em \vspace{-0.5em} % Compress items in list together for aesthetics
  }{
  \end{list}
  \vspace{0.5em} % Some space after the list of bullet points
}

% The below commands define the whitespace after certain things in the document - they can be \smallskip, \medskip or \bigskip
\def\namesize{\huge} % Size of the name at the top of the document
\def\addressskip{\smallskip} % The space between the two address (or phone/email) lines
\def\sectionlineskip{\medskip} % The space above the horizontal line for each section 
\def\nameskip{\bigskip} % The space after your name at the top
\def\sectionskip{\medskip} % The space after the heading section

"""

resume_class_type = DocumentClassType('resume', RESUME_DEFINITION)



from copy import deepcopy
from typing import Union, AnyStr, List, Optional

import numpy as np
import pandas as pd

from mixins.repr import ReprMixin

from pyexlatex.table.logic.panels.combine import (
    common_column_labels,
    common_row_labels,
    remove_label_collections_from_grid
)
from pyexlatex.table.logic.panels.letters import panel_string
from pyexlatex.table.logic.panels.topleft import _set_top_left_corner_labels
from pyexlatex.table.models.labels.table import LabelTable, LabelCollection
from pyexlatex.table.models.panels.panel import Panel
from pyexlatex.table.models.panels.panel import PanelGrid, GridShape
from pyexlatex.table.models.spacing.columntable import ColumnPadTable
from pyexlatex.table.models.spacing.rowtable import RowPadTable
from pyexlatex.table.models.table.section import TableSection
from pyexlatex.table.models.data.table import DataTable


class PanelCollection(ReprMixin):
    """
    Lays panel contents in a grid, consolidates labels, creates padding between tables

    """
    repr_cols = ['name', 'panels']
    _num_columns: int

    def __init__(self, panels: List[Panel], label_consolidation: str='object', enforce_label_order=True,
                 top_left_corner_labels: Union[LabelTable, LabelCollection, List[AnyStr], AnyStr]=None,
                 pad_rows: int=1, pad_columns: int=1, name: str=None):
        """

        :param panels: list of Panels, which represent a full set of rows of the table. for multiple
                        sections in one set of rows, create DataTables for each section and pass to Panels.
        :param label_consolidation: pass 'object' to compare object equality for label consolidation, 'str'
                                    for converting all labels to strings then comparing equality. Use 'object'
                                    for more control over consolidation.
        :param enforce_label_order: pass False to allow consolidating lower labels even if upper labels do not match.
                                    e.g. if labels on one table are [['Top1'], ['Bot1', 'Bot2']], then labels on the other
                                    table are [['Top2'], ['Bot1', 'Bot2']], consolidated labels when passing False will be
                                    ['Bot1', 'Bot2'], while when passing True, no labels will be consolidated. Under True,
                                    will start from the top label, then stop consolidating once it has a mismatch.
        :param top_left_corner_labels: additional labels to place in the top left corner. pass a single string
                                       or a list of strings for convenience. a list of strings will be create labels
                                       which span the gap horizontally and go downwards, one label per row. pass
                                       LabelCollection or LabelTable for more control.
        :param pad_rows: horizontal spacing to put between panels
        :param pad_columns: vertical spacing to put between TableSections
        :param name: name that will be used to construct caption in output
        """
        self.name = name
        self.panels = panels
        self.label_consolidation = label_consolidation.lower().strip() \
            if isinstance(label_consolidation, str) else label_consolidation
        self.enforce_label_order = enforce_label_order
        self.top_left_corner_labels = _set_top_left_corner_labels(top_left_corner_labels)
        self.pad_rows = pad_rows
        self.pad_columns = pad_columns

        self.has_row_labels = False
        self.has_column_labels = False

        self.consolidate_labels()
        self.pad_grid()

    def iterpanels(self, add_panel_order_label: bool = True):
        """
        First panel is headers. Then each original panel

        self.grid includes all panels as well as labels. Need to separate back out to
        get each panel

        :param add_panel_order_label: Whether to add to names of panels Panel A:, Panel B:,
        and so on

        :return:
        :rtype:
        """

        # Get lengths of panels
        if len(self.panels) > 1:
            # Include spacing which is added in pad_grid
            panel_lengths = [panel.panel_grid.shape[0] * 2 - 1 for panel in self.panels]
        else:
            panel_lengths = [self.panels[0].panel_grid.shape[0]]
        if self.has_column_labels:
            # First panel will have an additional row at the top if there are column labels
            panel_lengths[0] += 1

        total_row_idx = 0
        for panel_idx, panel_length in enumerate(panel_lengths):
            for row_idx in range(panel_length):
                row = self.rows[total_row_idx]
                # if first part of actual panel, below labels, and panel has a name
                if (
                        (
                            # If it's the first panel, and there's column labels, then second row should show name
                            self.has_column_labels and panel_idx == 0 and row_idx == 1 or
                            # If it's a later panel, or there's no column labels, first row should show name
                            (not self.has_column_labels or panel_idx > 0) and row_idx == 0
                        ) and
                        # Has to have name to show name
                        self.panels[panel_idx].name is not None
                ):
                    # output with the name
                    if add_panel_order_label:
                        full_name = panel_string(panel_idx) + self.panels[panel_idx].name
                    else:
                        full_name = self.panels[panel_idx].name
                    yield Panel(PanelGrid([row]), name=full_name)
                else:
                    # column labels panel, no matching name. Or no name for user supplied panel
                    yield Panel(PanelGrid([row]))
                total_row_idx += 1
            # If we are in-between panels, not on the last panel
            if panel_idx + 1 != len(self.panels) and self.pad_rows > 0:
                for _ in range(self.pad_rows):
                    # In between panels, there are also spacers, yield those
                    row = self.rows[total_row_idx]
                    yield Panel(PanelGrid([row]))
                    total_row_idx += 1

    @property
    def rows(self):
        try:
            return self._rows
        except AttributeError:
            self._rows = self._create_panel_rows()

        return self._rows

    @property
    def num_columns(self) -> int:
        try:
            return self._num_columns
        except AttributeError:
            self._num_columns = max([row.num_columns for row in self.rows])

        return self._num_columns

    def _create_panel_rows(self):
        rows: [TableSection] = []
        for panel_row in self.grid:
            new_row = None
            for i, section in enumerate(panel_row):
                if i == 0:
                    new_row = section
                else:
                    new_row = new_row + section
            if new_row:
                rows.append(new_row)

        try:
            num_columns = self._num_columns
        except AttributeError:
            num_columns = max([row.num_columns for row in rows])
        # Now pad rows
        for row in rows:
            row.pad(num_columns, direction='right')


        return rows


    @property
    def grid(self):
        try:
            return self._grid
        except AttributeError:
            self._grid = _concatenate_uneven_rows_filling_right(
                [panel.panel_grid for panel in self.panels],
                fill_value=TableSection([]),
                array_class=GridShape
            )

        return self._grid

    def consolidate_labels(self):

        # TODO [#46]: reduce complexity and improve testing of label consolidation
        #
        # This code has grown quite a bit to handle all the cases, and it
        # still has not been thoroughly tested for every case. Need to expand
        # the testing to cover all the cases and then try to simplify the code

        if self.label_consolidation is None:
            return

        if self.label_consolidation == 'object':
            use_object_equality = True
        elif self.label_consolidation in ('string', 'str', True):
            use_object_equality = False
        else:
            raise ValueError(f'must pass object, string, or None to label consolidation. Got {self.label_consolidation}')

        column_labels: [LabelTable] = common_column_labels(
            self.grid,
            use_object_equality=use_object_equality,
            enforce_label_order=self.enforce_label_order
        )
        # Column labels may be modified to add top left corner label, but need to
        # track original so it can be removed from existing panels
        orig_column_labels = deepcopy(column_labels)

        row_labels: [LabelTable] = common_row_labels(
            self.grid,
            use_object_equality=use_object_equality,
            enforce_label_order=self.enforce_label_order
        )

        # Remove from the original tables the labels that were just consolidated
        remove_label_collections_from_grid(
            self.grid,
            column_labels=orig_column_labels,
            row_labels=row_labels,
            use_object_equality=use_object_equality
        )

        if column_labels is not None:
            if row_labels is None and not self.top_left_corner_labels.is_spacer:
                # If there are column labels but not row labels, still need to deal with top left label.
                # Adding the top left label is handled in the if row_labels is not None block, but it will
                # not be reached as row_labels is None. Therefore create a blank label table for each grid row
                # except for the column labels. The top left label to go with the column row will be added
                # in the following block.
                # Grid shape -1 to exclude just added column labels
                row_labels: List[LabelTable] = []
                for grid_row in self.grid:
                    left_value: TableSection = grid_row[0]
                    if left_value in column_labels:
                        # top left corner label, handled separately in row labels section
                        continue
                    section_height = len(left_value.rows)
                    if isinstance(left_value, DataTable):
                        if left_value.column_labels in column_labels:
                            # Adjust for when the column labels have been consolidated into the overall column labels,
                            # then don't need to add row label as this column label won't be in the output
                            section_height -= 1
                    label_lol = [['']] * section_height
                    row_labels.append(LabelTable.from_list_of_lists(label_lol))

        column_labels_created = False
        if row_labels is not None:
            if column_labels is not None:
                self._add_column_labels(column_labels)
                column_labels_created = True
            # After adding column labels, there is an additional row at the top of the grid
            # Therefore we will need one additional LabelTable for the first row, which is the row of column labels
            # If top_left_corner_labels was passed on object creation, use that as LabelTable. Otherwise use a blank one
            if self.has_column_labels:
                all_row_labels = [self.top_left_corner_labels] + row_labels
            else:
                all_row_labels = row_labels
            self._add_row_labels(all_row_labels)
        elif column_labels is not None:
            # There are no common row labels, but if there are any row labels which are not being consolidated,
            # still need to add the top left corner
            # First detect the existence of any row labels
            any_row_labels = False
            for grid_row in self.grid:
                left_value: TableSection = grid_row[0]
                if isinstance(left_value, DataTable) and left_value.has_row_labels:
                    any_row_labels = True
            if any_row_labels:
                # Now add the top left corner
                column_labels[0] = self.top_left_corner_labels + column_labels[0]

        if column_labels is not None and not column_labels_created:
            self._add_column_labels(column_labels)


    def pad_grid(self):
        row_pad = RowPadTable()

        grid_rows: [GridShape] = []
        for n_row, grid_row in enumerate(self.grid):
            # Add first elem
            new_row = np.array([grid_row[0]]).view(GridShape).reshape(1,1)
            # Add following elems
            for n_elem, elem in enumerate(grid_row[1:]):
                # Add pads between following elems
                if self.pad_columns and not (n_elem == 0 and self.has_row_labels): # only skip first if there are row labels
                    new_row = np.append(new_row, np.array([ColumnPadTable(self.pad_columns)])).view(GridShape)
                new_row = np.append(new_row, elem).view(GridShape)
                new_row = new_row.reshape((1, new_row.shape[0])) # reorganize into row
            grid_rows.append(new_row)
            # add row padding on every loop except last
            if n_row != self.grid.shape[0] - 1:
                # no need to add padding after column labels
                if n_row == 0 and self.has_column_labels:
                    continue
                for i in range(self.pad_rows):
                    grid_rows.append(
                        np.array([row_pad]).view(GridShape).reshape(1,1)
                    )

        # Before combining rows, must have same number of elements. Pad right with empty label tables
        num_grid_columns = max(row.shape[1] for row in grid_rows) # find max number of columns
        out_grid_rows = []
        for row in grid_rows:
            pad_number = num_grid_columns - row.shape[1]
            assert pad_number >= 0
            new_row = np.append(row, np.array([row_pad] * pad_number)).view(GridShape)
            new_row = new_row.reshape((1, new_row.shape[0]))  # reorganize into row
            out_grid_rows.append(new_row)

        new_grid = np.concatenate(out_grid_rows).view(GridShape)
        self._grid = new_grid
        self._rows = self._create_panel_rows() # need to recreate rows with new grid

    def _add_column_labels(self, column_labels: List[LabelTable]):
        assert len(column_labels) == self.grid.shape[1]

        if all(table.is_empty for table in column_labels):
            # if no consolidated labels, no need to add
            self.has_column_labels = False
            return

        self.has_column_labels = True

        # Form PanelGrid from labels
        column_label_grid = PanelGrid(column_labels, shape=(1,len(column_labels)))

        # Combine label PanelGrid and existing PanelGrid
        self._grid = np.concatenate([column_label_grid, self._grid]).view(GridShape)

    def _add_row_labels(self, row_labels: List[LabelTable]):
        assert len(row_labels) == self.grid.shape[0]

        if all(table.is_empty for table in row_labels):
            # if no consolidated labels, no need to add
            self.has_row_labels = False
            return

        self.has_row_labels = True

        # Form PanelGrid from labels
        row_label_grid = PanelGrid(row_labels, shape=(len(row_labels), 1))

        # Combine label PanelGrid and existing PanelGrid
        self._grid = np.concatenate([row_label_grid, self._grid], axis=1).view(GridShape)

    @classmethod
    def from_list_of_lists_of_dfs(cls, df_list_of_lists: List[List[pd.DataFrame]],
                                  panel_names: List[str] = None, *args,
                                  panel_kwargs={},
                                  data_table_kwargs={}, **kwargs):
        """
        To create a single panel table, pass a single list within
        a list of DataFrames, e.g. [[df1, df2]] then shape will specify how the DataFrames will
        be organized in the Panel. If you pass two lists within the outer list, then shape will
        apply to each Panel. So [[df1, df2], [df3, df4]] with shape=(1,2) create a two Panel table
        with two tables placed within each panel going horizontally, so that the overall shape is (2,2).

        Note: convenience method for if not much control over table is needed.
        To apply different options to each panel, construct them individually using
        Panel.from_df_list

        :param df_list_of_lists:
        :param panel_names: list of panel names. Must be of same length as outer list in df_list_of_lists
        :param args: args to pass to PanelCollection constructor
        :param panel_kwargs: Panel.from_df_list kwargs. Same kwargs will be passed to all panels
        :param kwargs: kwargs to pass to PanelCollection constructor
        :param data_table_kwargs: kwargs to be passed to DataTable.from_df. Same kwargs will be passed to
                                  all data tables.

        :return: PanelCollection
        """
        _validate_panel_names(panel_names, df_list_of_lists)

        panels = []
        for i, df_list in enumerate(df_list_of_lists):
            panel_name = _panel_name_or_none(panel_names, i)
            panels.append(
                Panel.from_df_list(
                    df_list,
                    name=panel_name,
                    data_table_kwargs=data_table_kwargs,
                    **panel_kwargs)
            )

        label_consolidation = kwargs.pop('label_consolidation', 'string')

        return cls(
            panels,
            label_consolidation=label_consolidation,
            **kwargs
        )

    def to_tex(self, mid_rule=True):
        from pyexlatex.table.logic.table.build import build_tabular_content_from_panel_collection
        return build_tabular_content_from_panel_collection(self, mid_rule=mid_rule)

def _panel_name_or_none(panel_names: Optional[List[str]], index: int):
    if panel_names is not None:
        return panel_names[index]
    else:
        return None

def _validate_panel_names(panel_names: Optional[List[str]], df_list_of_lists: List[List[pd.DataFrame]]):
    if panel_names is None:
        return

    num_panel_names = len(panel_names)
    num_panels = len(df_list_of_lists)
    if num_panel_names != num_panels:
        raise ValueError(f'must pass as many panel names as panels. Got {num_panel_names} names '
                         f'and {num_panels} panels.')


def _concatenate_uneven_rows_filling_right(rows, fill_value=np.nan, array_class=None):
    """
    Concatenates along vertical axis, filling right as needed.

    Examples:
        a = np.array(
            [[1, 2, 3]]
        )
        b = np.array(
            [[4, 5]]
        )
        _concatenate_uneven_rows_filling_right([a, b])

        array([[ 1.,  2.,  3.],
               [ 4.,  5., nan]])

    Args:
        rows:
        fill_value:
        array_class:

    Returns:

    """
    max_len = max([row.shape[1] for row in rows])

    concat_rows = []
    for row in rows:
        num_to_add = max_len - row.shape[1]
        if num_to_add > 0:
            add_array = np.array([[fill_value] * num_to_add])
            concat_array = np.concatenate([row, add_array], axis=1)
        else:
            concat_array = row
        concat_rows.append(concat_array)

    out_arr = np.concatenate(concat_rows)

    if array_class:
        out_arr = out_arr.view(array_class)

    return out_arr


from pyexlatex.models.item import SimpleItem
from pyexlatex.models.section.base import TextAreaMixin
from pyexlatex.typing import PyexlatexItems


class Closing(TextAreaMixin, SimpleItem):
    name = 'closing'

    def __init__(self, closing: PyexlatexItems):
        self.closing = closing
        super().__init__(self.name, self.closing)



from typing import Optional, List, Dict, Type
from data import Data
from latex.exc import LatexBuildError

from pyexlatex.logic.output.api.builders.base import BaseBuilder
from pyexlatex.logic.output.api.builders.htlatex import HTLatexBuilder
from pyexlatex.logic.output.api.formats import OutputFormats
from pyexlatex.logic.output.errors.exc import (
    exception_manager
)
from pyexlatex.logic.output.api.exc_handler.main import APIExceptionHandler
from pyexlatex.logic.output.api.exc_handler.prepend.typing import PrependKwargsDict, PrependItemsDict
from pyexlatex.logic.output.api.exc_handler.prepend.main import add_prepend_items_dict_to_latex_str
from pyexlatex.logic.output.api.builders.lualatex import LuaLatexBuilder

BUILDERS: Dict[OutputFormats, Type[BaseBuilder]] = {
    OutputFormats.PDF: LuaLatexBuilder,
    OutputFormats.HTML: HTLatexBuilder,
}


def latex_str_to_obj(latex_str: str, output_format: OutputFormats = OutputFormats.PDF,
                     texinputs: Optional[List[str]] = None, run_bibtex: bool = False,
                     retries_remaining: int = 3,
                     prepend_items_dict: PrependItemsDict = None,
                     prepend_kwargs_dict: PrependKwargsDict = None) -> Data:
    try:
        new_latex_str = add_prepend_items_dict_to_latex_str(prepend_items_dict, latex_str)
        return _latex_to_obj(new_latex_str, output_format, texinputs=texinputs, run_bibtex=run_bibtex)
    except LatexBuildError as e:
        exceptions = exception_manager.exceptions_from_latex_build_error(e)
        handler = APIExceptionHandler(
            exceptions,
            e,
            latex_str,
            prepend_kwargs_dict=prepend_kwargs_dict,
            prepend_items_dict=prepend_items_dict,
            retries_remaining=retries_remaining,
            texinputs=texinputs,
            run_bibtex=run_bibtex,
            output_format=output_format,
        )
        return handler.handle_exceptions()


def _latex_to_obj(latex_str: str, output_format: OutputFormats, texinputs: Optional[List[str]] = None,
                  run_bibtex: bool = False) -> Data:
    if texinputs is None:
        texinputs = []
    builder = BUILDERS[output_format]
    obj = builder().build(latex_str, texinputs=texinputs, run_bibtex=run_bibtex)
    return obj





import datetime

import pyexlatex as pl
from pyexlatex import Package, LeftFooter, RightFooter, FooterLine, CenterFooter, ThisPageNumber
from pyexlatex.models.blindtext import BlindText
from pyexlatex.models.document import Document
from pyexlatex.models.page.number import PageReference
from pyexlatex.texgen.packages.default import default_packages
from tests.base import INPUT_FILES_DIR, GENERATED_FILES_DIR
from tests.utils.pdf import compare_pdfs


class TestDocument:
    no_options_result = '\\documentclass[]{article}\n\\usepackage{amsmath}\n\\usepackage{pdflscape}\n\\usepackage{booktabs}\n\\usepackage{array}\n\\usepackage{threeparttable}\n\\usepackage{fancyhdr}\n\\usepackage{lastpage}\n\\usepackage{textcomp}\n\\usepackage{dcolumn}\n\\newcolumntype{L}[1]{>{\\raggedright\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{C}[1]{>{\\centering\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{R}[1]{>{\\raggedleft\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{.}{D{.}{.}{-1}}\n\\usepackage[T1]{fontenc}\n\\usepackage{caption}\n\\usepackage{subcaption}\n\\usepackage{graphicx}\n\\usepackage[margin=0.8in, bottom=1.2in]{geometry}\n\\usepackage[page]{appendix}\n\\pagestyle{fancy}\n\\renewcommand{\\headrulewidth}{0pt}\n\\fancyhead{}\n\\rfoot{Page \\thepage\\  of \\pageref{LastPage}}\n\\cfoot{}\n\\begin{document}\nwoo\n\\end{document}'
    common_kwargs = dict(
        packages=default_packages + ['hyperref', Package('tikz')],
        landscape=False,  # TODO: set landscape to True in Document tests once landscape bug is fixed
        title='My Title',
        authors=['Person One', 'Human Two'],
        date=datetime.datetime(2020, 10, 1),
        abstract='Some text for the abstract',
        skip_title_page=False,
        page_modifier_str='margin=1in',
        page_header=True,
        page_numbers=False,
        appendix_modifier_str='page',
        section_numbering_styles=dict(
            section=r'\Roman{section}',
            subsection=r'\thesection.\Alph{subsection}',
            subsubsection=r'\thesubsection.\arabic{subsubsection}',
            subfigure=r'\roman{subfigure}',
        ),
        floats_at_end=True,
        floats_at_end_options='nolists',
        document_type='article',
        font_size=12,
        num_columns=2,
        line_spacing=2,
        tables_relative_font_size=-1,
        figures_relative_font_size=1,
        page_style='fancy',
        custom_headers=[
            pl.Header(pl.SmallCaps('short title'), align='left'),
            pl.Header(pl.SmallCaps(['Page ', pl.ThisPageNumber()]))
        ],
        remove_section_numbering=True,
        separate_abstract_page=True,
        extra_title_page_lines=[pl.Italics('Stuff:'), 'Yeah, things'],
        empty_title_page_style=False,
        pre_output_func=lambda x: x + '\n\n\n',
    )

    def test_no_options_str(self):
        doc = Document('woo')
        assert str(doc) == self.no_options_result

    def test_no_options_list(self):
        doc = Document(['woo'])
        assert str(doc) == self.no_options_result

    def test_basic_title_page(self):
        doc = Document([BlindText(), ''] * 10, authors=['Person One', 'Human Two', 'Being Three'], title='My Title')
        assert str(doc) == '\\documentclass[]{article}\n\\usepackage{blindtext}\n\\usepackage{amsmath}\n\\usepackage{pdflscape}\n\\usepackage{booktabs}\n\\usepackage{array}\n\\usepackage{threeparttable}\n\\usepackage{fancyhdr}\n\\usepackage{lastpage}\n\\usepackage{textcomp}\n\\usepackage{dcolumn}\n\\newcolumntype{L}[1]{>{\\raggedright\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{C}[1]{>{\\centering\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{R}[1]{>{\\raggedleft\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{.}{D{.}{.}{-1}}\n\\usepackage[T1]{fontenc}\n\\usepackage{caption}\n\\usepackage{subcaption}\n\\usepackage{graphicx}\n\\usepackage[margin=0.8in, bottom=1.2in]{geometry}\n\\usepackage[page]{appendix}\n\\pagestyle{fancy}\n\\renewcommand{\\headrulewidth}{0pt}\n\\fancyhead{}\n\\rfoot{Page \\thepage\\  of \\pageref{LastPage}}\n\\cfoot{}\n\\begin{document}\n\\title{My Title}\n\\author{Person One, Human Two, and Being Three}\n\\date{\\today}\n\\maketitle\n\\blindtext\n\n\\blindtext\n\n\\blindtext\n\n\\blindtext\n\n\\blindtext\n\n\\blindtext\n\n\\blindtext\n\n\\blindtext\n\n\\blindtext\n\n\\blindtext\n\n\\end{document}'

    def test_title_page_with_consistent_formatting(self):
        doc = Document([BlindText(), ''] * 10, authors=['Person One', 'Human Two', 'Being Three'], title='My Title', apply_page_style_to_section_starts=True)
        assert str(doc) == '\\documentclass[]{article}\n\\usepackage{blindtext}\n\\usepackage{amsmath}\n\\usepackage{pdflscape}\n\\usepackage{booktabs}\n\\usepackage{array}\n\\usepackage{threeparttable}\n\\usepackage{fancyhdr}\n\\usepackage{lastpage}\n\\usepackage{textcomp}\n\\usepackage{dcolumn}\n\\newcolumntype{L}[1]{>{\\raggedright\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{C}[1]{>{\\centering\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{R}[1]{>{\\raggedleft\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{.}{D{.}{.}{-1}}\n\\usepackage[T1]{fontenc}\n\\usepackage{caption}\n\\usepackage{subcaption}\n\\usepackage{graphicx}\n\\usepackage[margin=0.8in, bottom=1.2in]{geometry}\n\\usepackage[page]{appendix}\n\\pagestyle{fancy}\n\\renewcommand{\\headrulewidth}{0pt}\n\\fancyhead{}\n\\rfoot{Page \\thepage\\  of \\pageref{LastPage}}\n\\cfoot{}\n\\fancypagestyle{plain}{\\rfoot{Page \\thepage\\  of \\pageref{LastPage}}\n\\cfoot{}}\n\\begin{document}\n\\title{My Title}\n\\author{Person One, Human Two, and Being Three}\n\\date{\\today}\n\\maketitle\n\\blindtext\n\n\\blindtext\n\n\\blindtext\n\n\\blindtext\n\n\\blindtext\n\n\\blindtext\n\n\\blindtext\n\n\\blindtext\n\n\\blindtext\n\n\\blindtext\n\n\\end{document}'

    def test_custom_footer(self):
        doc = Document(
            'woo',
            custom_footers=[
                LeftFooter('stuff'),
                CenterFooter(''),
                RightFooter(
                    ['Page', ThisPageNumber(), '\\', 'of', PageReference('LastPage')]
                ),
                FooterLine()
            ]
        )
        assert str(doc) == '\\documentclass[]{article}\n\\usepackage{amsmath}\n\\usepackage{pdflscape}\n\\usepackage{booktabs}\n\\usepackage{array}\n\\usepackage{threeparttable}\n\\usepackage{fancyhdr}\n\\usepackage{lastpage}\n\\usepackage{textcomp}\n\\usepackage{dcolumn}\n\\newcolumntype{L}[1]{>{\\raggedright\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{C}[1]{>{\\centering\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{R}[1]{>{\\raggedleft\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{.}{D{.}{.}{-1}}\n\\usepackage[T1]{fontenc}\n\\usepackage{caption}\n\\usepackage{subcaption}\n\\usepackage{graphicx}\n\\usepackage[margin=0.8in, bottom=1.2in]{geometry}\n\\usepackage[page]{appendix}\n\\pagestyle{fancy}\n\\renewcommand{\\headrulewidth}{0pt}\n\\fancyhead{}\n\\lfoot{stuff}\n\\cfoot{}\n\\rfoot{Page\n\\thepage\n\\\nof\n\\pageref{LastPage}}\n\\renewcommand{\\footrulewidth}{0.4pt}\n\\begin{document}\nwoo\n\\end{document}'

    def test_all_options_with_title_page(self):
        doc = Document(
            'woo',
            **self.common_kwargs,
        )

        assert str(doc) == '\\documentclass[12pt, twocolumn]{article}\n\\usepackage{amsmath}\n\\usepackage{pdflscape}\n\\usepackage{booktabs}\n\\usepackage{array}\n\\usepackage{threeparttable}\n\\usepackage{fancyhdr}\n\\usepackage{lastpage}\n\\usepackage{textcomp}\n\\usepackage{dcolumn}\n\\newcolumntype{L}[1]{>{\\raggedright\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{C}[1]{>{\\centering\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{R}[1]{>{\\raggedleft\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{.}{D{.}{.}{-1}}\n\\usepackage[T1]{fontenc}\n\\usepackage{caption}\n\\usepackage{subcaption}\n\\usepackage{graphicx}\n\\usepackage{hyperref}\n\\usepackage{tikz}\n\\usepackage[margin=1in]{geometry}\n\\usepackage{floatrow}\n\\DeclareFloatFont{small}{\\small}\n\\floatsetup[table]{font=small,capposition=top}\n\\DeclareFloatFont{large}{\\large}\n\\floatsetup[figure]{font=large,capposition=top}\n\\usepackage[nolists]{endfloat}\n\\DeclareDelayedFloatFlavor{ltable}{table}\n\\DeclareDelayedFloatFlavor{lfigure}{figure}\n\\usepackage{setspace}\n\\doublespacing\n\\usepackage[page]{appendix}\n\\renewcommand{\\thesection}{\\Roman{section}}\n\\renewcommand{\\thesubsection}{\\thesection.\\Alph{subsection}}\n\\renewcommand{\\thesubsubsection}{\\thesubsection.\\arabic{subsubsection}}\n\\renewcommand{\\thesubfigure}{\\roman{subfigure}}\n\\setcounter{secnumdepth}{0}\n\\pagestyle{fancy}\n\\lhead{\\textsc{short title}}\n\\rhead{\\textsc{Page \n\\thepage}}\n\\cfoot{}\n\\begin{document}\n\\title{My Title}\n\\author{Person One and Human Two}\n\\date{2020-10-01 00:00:00}\n\\maketitle\n\\newpage\n\\thispagestyle{empty}\n{\n\\begin{center}\n\\setlength{\\baselineskip}{40pt}\n\\LARGE\n\\textbf{My Title}\n\\end{center}\n}\n\\vspace{35pt}\n\\begin{center}\n\\textbf{Abstract}\n\\end{center}\n\\vspace{-10pt}\n\\hspace{1.5em}\nSome text for the abstract\n\\vspace{20pt}\n\\textit{Stuff:}\n\n\nYeah, things\n\n\n\\newpage\n\\setcounter{page}{1}\nwoo\n\n\n\n\\end{document}'

    def test_all_options_no_title_page(self):
        kwargs = {**self.common_kwargs}
        kwargs['skip_title_page'] = True

        doc = Document(
            'woo',
            **kwargs
        )

        assert str(doc) == '\\documentclass[12pt, twocolumn]{article}\n\\usepackage{amsmath}\n\\usepackage{pdflscape}\n\\usepackage{booktabs}\n\\usepackage{array}\n\\usepackage{threeparttable}\n\\usepackage{fancyhdr}\n\\usepackage{lastpage}\n\\usepackage{textcomp}\n\\usepackage{dcolumn}\n\\newcolumntype{L}[1]{>{\\raggedright\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{C}[1]{>{\\centering\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{R}[1]{>{\\raggedleft\\let\\newline\\\\\\arraybackslash\\hspace{0pt}}m{#1}}\n\\newcolumntype{.}{D{.}{.}{-1}}\n\\usepackage[T1]{fontenc}\n\\usepackage{caption}\n\\usepackage{subcaption}\n\\usepackage{graphicx}\n\\usepackage{hyperref}\n\\usepackage{tikz}\n\\usepackage[margin=1in]{geometry}\n\\usepackage{floatrow}\n\\DeclareFloatFont{small}{\\small}\n\\floatsetup[table]{font=small,capposition=top}\n\\DeclareFloatFont{large}{\\large}\n\\floatsetup[figure]{font=large,capposition=top}\n\\usepackage[nolists]{endfloat}\n\\DeclareDelayedFloatFlavor{ltable}{table}\n\\DeclareDelayedFloatFlavor{lfigure}{figure}\n\\usepackage{setspace}\n\\doublespacing\n\\usepackage[page]{appendix}\n\\renewcommand{\\thesection}{\\Roman{section}}\n\\renewcommand{\\thesubsection}{\\thesection.\\Alph{subsection}}\n\\renewcommand{\\thesubsubsection}{\\thesubsection.\\arabic{subsubsection}}\n\\renewcommand{\\thesubfigure}{\\roman{subfigure}}\n\\setcounter{secnumdepth}{0}\n\\pagestyle{fancy}\n\\lhead{\\textsc{short title}}\n\\rhead{\\textsc{Page \n\\thepage}}\n\\cfoot{}\n\\begin{document}\nwoo\n\n\n\n\\end{document}'

    def test_to_pdf(self):
        doc = Document('woo')
        doc.to_pdf(GENERATED_FILES_DIR)
        compare_pdfs(INPUT_FILES_DIR / 'document.pdf', GENERATED_FILES_DIR / 'document.pdf')

    def test_to_html(self):
        doc = Document('woo')
        doc.to_html(GENERATED_FILES_DIR)
        expect_str = (INPUT_FILES_DIR / 'document.html').read_text()
        generated_str = (GENERATED_FILES_DIR / 'document.html').read_text()
        assert _remove_meta_tags(expect_str) == _remove_meta_tags(generated_str)


def _remove_meta_tags(content: str) -> str:
    lines = content.splitlines()
    out_str = ''
    for line in lines:
        if line.startswith('<meta'):
            continue
        else:
            out_str += line
            out_str += '\n'
    return out_str



from typing import Optional, Sequence, Union, List, Any
import pyexlatex as pl
from pyexlatex.models.template import Template
from pyexlatex.models.format.text.bold import Bold
from pyexlatex.models.hyperlinks import Hyperlink
from pyexlatex.logic.format.and_join import join_with_commas_and_and_output_list
from pyexlatex.models.format.hangindent import HangIndent
from pyexlatex.models.format.paragraph.justifying import Justifying


class Publication(Template):
    SPACE_BETWEEN_ADJUSTMENT = 0.2

    def __init__(self, title: str, co_authors: Optional[Sequence[str]] = None, journal_info: Optional[str] = None,
                 href: Optional[str] = None, extra_info: Optional[str] = None, prevent_page_break: bool = True,
                 description: Optional[Any] = None):
        self.title = title
        self.co_authors = co_authors
        self.journal_info = journal_info
        self.href = href
        self.extra_info = extra_info
        self.prevent_page_break = prevent_page_break
        self.description = description

        self.contents = self._get_contents()
        super().__init__()

    def _get_contents(self):
        possible_items = [
            HangIndent(1),
            self._title_str,
            pl.Italics(self.journal_info) if self.journal_info is not None else None,
            self._with_coauthors,
            self.extra_info,
            *self._description_content,
            pl.VSpace(self.SPACE_BETWEEN_ADJUSTMENT),
        ]
        items = [item for item in possible_items if item is not None]
        if self.prevent_page_break:
            items = pl.NoPageBreak(items)
        return items

    @property
    def _title_obj(self) -> Union[Bold, Hyperlink]:
        if self.href is None:
            title = Bold(self.title)
        else:
            # Handle hyperlink
            styled_title = pl.TextColor(self.title, color=pl.RGB(50, 82, 209, color_name='darkblue'))
            styled_title = pl.Underline(styled_title)
            title = Hyperlink(self.href, styled_title)
            self.add_data_from_content(title)
        return title

    @property
    def _title_str(self) -> str:
        title_str = str(self._title_obj)
        if any([item is not None for item in [self.co_authors, self.journal_info, self.extra_info]]):
            title_str += ','
        return title_str


    @property
    def _with_coauthors(self) -> Optional[List[str]]:
        if self.co_authors is None:
            return None

        return [' with ', *join_with_commas_and_and_output_list(self.co_authors)]

    @property
    def _description_content(self) -> List:
        if self.description is None:
            return []

        return [
            '',  # separate paragraph
            Justifying(),  # return to justifying on both sides
            self.description,
            pl.VSpace(0.2)  # extra spacing looks good when description is included
        ]

